# -*- coding: utf-8 -*-
#
#   Finite State Machine
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import weakref
from abc import ABC
from typing import List, Optional, Dict, Generic

from .machine import S, C, U, T
from .machine import Transition, State, Machine, Status, Delegate


class BaseTransition(Transition, ABC, Generic[C]):
    """ Transition with the name of target state """

    def __init__(self, target: str):
        super().__init__()
        self.__target = target

    @property
    def target(self) -> str:
        return self.__target


class BaseState(State, ABC, Generic[C, T]):
    """ State with transitions """

    def __init__(self):
        super().__init__()
        self.__transitions: List[T] = []

    def add_transition(self, transition: T):
        assert transition not in self.__transitions, 'transition exists'
        self.__transitions.append(transition)

    # Override
    def evaluate(self, ctx: C) -> Optional[T]:
        for trans in self.__transitions:
            if trans.evaluate(ctx):
                # OK, get target state from this transition
                return trans


class BaseMachine(Machine, Generic[C, T, S]):

    def __init__(self, default: str):
        super().__init__()
        self.__default = default
        self.__current: Optional[S] = None
        self.__states: Dict[str, S] = {}   # name(str) => State
        self.__delegate: Optional[weakref.ReferenceType] = None
        self.__status: Status = Status.STOPPED

    @property
    def delegate(self) -> Delegate[C, T, S]:
        if self.__delegate is not None:
            return self.__delegate()

    @delegate.setter
    def delegate(self, handler: Delegate[C, T, S]):
        if handler is None:
            self.__delegate = None
        else:
            self.__delegate = weakref.ref(handler)

    @property
    def context(self) -> C:
        raise NotImplemented

    #
    #   States
    #
    def add_state(self, name: str, state: S):
        self.__states[name] = state

    def get_state(self, name: str) -> S:
        return self.__states.get(name)

    @property  # Override
    def default_state(self) -> S:
        return self.__states.get(self.__default)

    @property  # Override
    def current_state(self) -> S:
        return self.__current

    @current_state.setter  # Override
    def current_state(self, state: S):
        self.__current = state

    # Override
    def target_state(self, transition: BaseTransition[C]) -> S:
        return self.__states.get(transition.target)

    # Override
    def change_state(self, state: Optional[S]):
        machine = self.context
        old = self.current_state
        delegate = self.delegate
        # events before state changed
        if delegate is not None:
            delegate.enter_state(state, machine)
        if state is not None:
            state.on_enter(machine)
        # change state
        self.current_state = state
        # events after state changed
        if delegate is not None:
            delegate.exit_state(old, machine)
        if old is not None:
            old.on_exit(machine)

    #
    #   Actions
    #

    # Override
    def start(self):
        self.change_state(state=self.default_state)
        self.__status = Status.RUNNING

    # Override
    def stop(self):
        self.__status = Status.STOPPED
        self.change_state(state=None)

    # Override
    def pause(self):
        machine = self.context
        current = self.current_state
        # events before state paused
        delegate = self.delegate
        if delegate is not None:
            delegate.pause_state(current, machine)
        current.on_pause(machine)
        # pause state
        self.__status = Status.PAUSED

    # Override
    def resume(self):
        machine = self.context
        current = self.current_state
        # reuse state
        self.__status = Status.RUNNING
        # events after state resumed
        delegate = self.delegate
        if delegate is not None:
            delegate.resume_state(current, machine)
        current.on_resume(machine)

    #
    #   Ticker
    #

    # Override
    def tick(self):
        machine = self.context
        current = self.current_state
        if current is not None and self.__status == Status.RUNNING:
            trans = current.evaluate(machine)
            if isinstance(trans, BaseTransition):
                target = self.target_state(transition=trans)
                assert target is not None, 'target state error: %s' % trans.target
                self.change_state(state=target)



# -*- coding: utf-8 -*-
#
#   TCP: Transmission Control Protocol
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import threading
import time
import weakref
from abc import abstractmethod
from typing import Optional, Set, Dict, MutableMapping

from ..fsm import Ticker

from .connection import Connection
from .hub import Hub


class BaseHub(Hub, Ticker):

    DYING_EXPIRES = 120  # kill connection after 2 minutes

    def __init__(self):
        super().__init__()
        self.__lock = threading.RLock()
        self.__connections: Set[Connection] = set()
        # because the remote address will always different to local address, so
        # we shared the same map for all directions here:
        #    mapping: (remote, local) => Connection
        #    mapping: (remote, null) => Connection
        #    mapping: (local, null) => Connection
        self.__map: Dict[tuple, MutableMapping[tuple, Connection]] = {}
        # mapping: (remote, local) => time to kill
        self.__dying_times: Dict[tuple, int] = {}

    # Override
    def send(self, data: bytes, source: Optional[tuple], destination: tuple) -> bool:
        conn = self.connect(remote=destination, local=source)
        if conn is None or not conn.opened:
            # connection closed
            return False
        return conn.send(data=data, target=destination) != -1

    # Override
    def get_connection(self, remote: Optional[tuple], local: Optional[tuple]) -> Optional[Connection]:
        with self.__lock:
            return self.__seek(remote=remote, local=local)

    # Override
    def connect(self, remote: Optional[tuple], local: Optional[tuple]) -> Optional[Connection]:
        assert local is not None or remote is not None, 'both local & remote addresses are empty'
        # 1. try to get connection from cache pool
        conn = self.get_connection(remote=remote, local=local)
        if conn is not None:
            return conn
        # 2. connection not found, try to create connection
        with self.__lock:
            # double check to make sure the connection doesn't exist
            conn = self.__seek(remote=remote, local=local)
            if conn is None:
                # create it
                conn = self.create_connection(remote=remote, local=local)
                if self.__create_indexes(connection=conn, remote=remote, local=local):
                    # make sure different connection with same pair(remote, local) not exists
                    self.__connections.discard(conn)
                    # cache it
                    self.__connections.add(conn)
        return conn

    @abstractmethod
    def create_connection(self, remote: Optional[tuple], local: Optional[tuple]) -> Connection:
        raise NotImplemented

    # Override
    def disconnect(self, remote: Optional[tuple], local: Optional[tuple]):
        with self.__lock:
            conn = self.__seek(remote=remote, local=local)
        if conn is not None:
            self.__remove(connection=conn)
            conn.close()

    # Override
    def tick(self):
        # call 'tick()' to drive all connections
        connections = set()
        with self.__lock:
            for conn in self.__connections:
                connections.add(conn)
        # check closed connection(s)
        now = int(time.time())
        for conn in connections:
            # call 'tick()' to drive all connections
            conn.tick()
            # check closed connection(s)
            pair = build_pair(remote=conn.remote_address, local=conn.local_address)
            if pair is None:
                # should not happen
                self.__remove(connection=conn)
            elif conn.opened:
                # connection still alive, revive it
                self.__dying_times.pop(pair, None)
            else:
                # connection is closed, check dying time
                expired = self.__dying_times.get(pair)
                if expired is None or expired == 0:
                    # set death clock
                    self.__dying_times[pair] = now + self.DYING_EXPIRES
                elif expired < now:
                    # times up, kill it
                    self.__remove(connection=conn)
                    # clear the death clock for it
                    self.__dying_times.pop(pair, None)

    def __seek(self, remote: Optional[tuple], local: Optional[tuple]) -> Optional[Connection]:
        if remote is None:
            assert local is not None, 'both local & remote addresses are empty'
            # get connection bound to local address
            table = self.__map.get(local)
            if table is None:
                return None
            else:
                # mapping: (local, null) => Connection
                conn = table.get(ANY_REMOTE_ADDRESS)
                if conn is not None:
                    return conn
            # get any connection bound to this local address
            for v in table.values():
                if v is not None:
                    return v
        else:
            # get connections connected to remote address
            table = self.__map.get(remote)
            if table is None:
                return None
            elif local is not None:
                # mapping: (remote, local) => Connection
                return table.get(local)
            else:
                # mapping: (remote, null) => Connection
                conn = table.get(ANY_LOCAL_ADDRESS)
                if conn is not None:
                    return conn
            # get any connection connected to this remote address
            for v in table.values():
                if v is not None:
                    return v

    def __create_indexes(self, connection: Connection, remote: Optional[tuple], local: Optional[tuple]) -> bool:
        if remote is None:
            if local is None:
                # raise ValueError('both local & remote addresses are empty')
                return False
            # get table for local address
            table = self.__map.get(local)
            if table is None:
                table = weakref.WeakValueDictionary()
                self.__map[local] = table
            # mapping: (local, null) => Connection
            table[ANY_REMOTE_ADDRESS] = connection
        else:
            # get table for remote address
            table = self.__map.get(remote)
            if table is None:
                table = weakref.WeakValueDictionary()
                self.__map[remote] = table
            if local is None:
                # mapping: (remote, null) => Connection
                table[ANY_LOCAL_ADDRESS] = connection
            else:
                # mapping: (remote, local) => Connection
                table[local] = connection
        return True

    def __remove_indexes(self, connection: Connection):
        remote = connection.remote_address
        local = connection.local_address
        if remote is None:
            if local is None:
                # raise ValueError('both local & remote addresses are empty')
                return
            # get table for local address
            table = self.__map.get(local)
            if table is not None:
                # mapping: (local, null) => Connection
                table.pop(ANY_REMOTE_ADDRESS, None)
        else:
            # get table for remote address
            table = self.__map.get(remote)
            if table is not None:
                if local is None:
                    # mapping: (remote, null) => Connection
                    table.pop(ANY_LOCAL_ADDRESS, None)
                else:
                    # mapping: (remote, local) => Connection
                    table.pop(local, None)

    def __remove(self, connection: Connection):
        with self.__lock:
            self.__remove_indexes(connection=connection)
            self.__connections.discard(connection)


def build_pair(remote: Optional[tuple], local: Optional[tuple]) -> Optional[tuple]:
    if remote is None:
        if local is None:
            # raise ValueError('both local & remote addresses are empty')
            return None
        return ANY_REMOTE_ADDRESS, local
    elif local is None:
        return remote, ANY_LOCAL_ADDRESS
    else:
        return remote, local


ANY_LOCAL_ADDRESS = ('0.0.0.0', 0)
ANY_REMOTE_ADDRESS = ('0.0.0.0', 0)



# -*- coding: utf-8 -*-
#
#   TCP: Transmission Control Protocol
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from .hub import Hub
from .channel import Channel
from .connection import Connection, Delegate as ConnectionDelegate

from .state import ConnectionState, StateMachine as ConnectionStateMachine

from .base_hub import BaseHub
from .base_channel import BaseChannel
from .base_conn import BaseConnection
from .active_conn import ActiveConnection

__all__ = [

    #
    #   Interfaces
    #
    'Hub',
    'Channel',
    'Connection', 'ConnectionDelegate',

    #
    #   FSM for Connection
    #
    'ConnectionState', 'ConnectionStateMachine',

    #
    #   Base
    #
    'BaseHub',
    'BaseChannel',
    'BaseConnection',
    'ActiveConnection',
]



# -*- coding: utf-8 -*-
#
#   Finite State Machine
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import time
from abc import ABC, abstractmethod
from threading import Thread
from typing import Optional, Generic

from .machine import S, C, U, T
from .base import BaseMachine


class Runnable(ABC):

    @abstractmethod
    def run(self):
        raise NotImplemented


class AutoMachine(BaseMachine, Runnable, ABC, Generic[C, T, S]):

    def __init__(self, default: str):
        super().__init__(default=default)
        self.__thread: Optional[Thread] = None

    # Override
    def start(self):
        super().start()
        if self.__thread is None:
            self.__thread = Thread(target=self.run)
            self.__thread.start()

    # Override
    def stop(self):
        super().stop()
        t: Thread = self.__thread
        if t is not None:
            t.join()
            self.__thread = None

    # Override
    def run(self):
        self.setup()
        try:
            self.handle()
        finally:
            self.finish()

    def setup(self):
        """ prepare for running """
        pass

    def finish(self):
        """ clean up after run """
        pass

    def handle(self):
        while self.current_state is not None:
            self.tick()
            self.idle()

    # noinspection PyMethodMayBeStatic
    def idle(self):
        time.sleep(0.125)



# -*- coding: utf-8 -*-
#
#   TCP: Transmission Control Protocol
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod
from typing import Optional


class Channel(ABC):

    @property
    def opened(self) -> bool:
        """ is_open() """
        raise NotImplemented

    @property
    def bound(self) -> bool:
        """ is_bound() """
        raise NotImplemented

    @abstractmethod
    def close(self):
        raise NotImplemented

    #
    #   Byte Channel
    #

    @abstractmethod
    def read(self, max_len: int) -> Optional[bytes]:
        raise NotImplemented

    @abstractmethod
    def write(self, data: bytes) -> int:
        raise NotImplemented

    #
    #   Selectable Channel
    #

    @abstractmethod
    def configure_blocking(self, blocking: bool):
        raise NotImplemented

    @property
    def blocking(self) -> bool:
        """ is_blocking() """
        raise NotImplemented

    #
    #   Network Channel
    #

    @abstractmethod
    def bind(self, address: Optional[tuple] = None, host: Optional[str] = '0.0.0.0', port: Optional[int] = 0):
        raise NotImplemented

    @property
    def local_address(self) -> Optional[tuple]:  # (str, int)
        raise NotImplemented

    #
    #   Socket/Datagram Channel
    #

    @property
    def connected(self) -> bool:
        """ is_connected() """
        raise NotImplemented

    @abstractmethod
    def connect(self, address: Optional[tuple] = None, host: Optional[str] = '127.0.0.1', port: Optional[int] = 0):
        raise NotImplemented

    @property
    def remote_address(self) -> Optional[tuple]:  # (str, int)
        raise NotImplemented

    #
    #   Datagram Channel
    #

    @abstractmethod
    def disconnect(self):
        raise NotImplemented

    @abstractmethod
    def receive(self, max_len: int) -> (Optional[bytes], Optional[tuple]):
        raise NotImplemented

    @abstractmethod
    def send(self, data: bytes, target: tuple) -> int:
        raise NotImplemented



# -*- coding: utf-8 -*-
#
#   TCP: Transmission Control Protocol
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import weakref
from abc import ABC, abstractmethod
from typing import Optional

from .net import Channel
from .net import Connection, ConnectionDelegate
from .net import BaseConnection, ActiveConnection
from .net import BaseHub


class StreamHub(BaseHub):
    """ Base Stream Hub """

    def __init__(self, delegate: ConnectionDelegate):
        super().__init__()
        self.__delegate = weakref.ref(delegate)

    @property
    def delegate(self) -> ConnectionDelegate:
        return self.__delegate()

    # Override
    def create_connection(self, remote: Optional[tuple], local: Optional[tuple]) -> Connection:
        # create connection with channel
        sock = self.create_channel(remote=remote, local=local)
        conn = BaseConnection(remote=remote, local=local, channel=sock)
        # set delegate
        if conn.delegate is None:
            conn.delegate = self.delegate
        # start FSM
        conn.start()
        return conn

    @abstractmethod
    def create_channel(self, remote: Optional[tuple], local: Optional[tuple]) -> Channel:
        raise NotImplemented


class ActiveStreamHub(StreamHub, ABC):
    """ Active Stream Hub """

    # Override
    def create_connection(self, remote: Optional[tuple], local: Optional[tuple]) -> Connection:
        # create connection with addresses
        conn = ActiveStreamConnection(remote=remote, local=local, hub=self)
        # set delegate
        if conn.delegate is None:
            conn.delegate = self.delegate
        # start FSM
        conn.start()
        return conn


class ActiveStreamConnection(ActiveConnection):
    """ Active Stream Connection """

    def __init__(self, remote: tuple, local: Optional[tuple], hub: ActiveStreamHub):
        super().__init__(remote=remote, local=local)
        self.__hub = weakref.ref(hub)

    # Override
    def connect(self, remote: tuple, local: Optional[tuple] = None) -> Channel:
        hub = self.__hub()
        # assert isinstance(hub, ActiveStreamHub)
        return hub.create_channel(remote=remote, local=local)



# -*- coding: utf-8 -*-
#
#   TCP: Transmission Control Protocol
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import socket
import threading
from abc import abstractmethod
from typing import Optional

from .channel import Channel
from .state import ConnectionState, StateMachine
from .base_conn import BaseConnection


class ActiveConnection(BaseConnection):

    def __init__(self, remote: tuple, local: Optional[tuple] = None, channel: Optional[Channel] = None):
        super().__init__(remote=remote, local=local, channel=channel)
        self.__lock = threading.RLock()
        self.__connecting = 0
        self.__running = False

    @abstractmethod
    def connect(self, remote: tuple, local: Optional[tuple] = None) -> Channel:
        raise NotImplemented

    def __reconnect(self):
        redo = False
        with self.__lock:
            try:
                self.__connecting += 1
                if self.__connecting == 1 and self.__running:
                    self.change_state(name=ConnectionState.CONNECTING)
                    self._channel = self.connect(remote=self._remote_address, local=self._local_address)
                    if self._channel is None:
                        self.change_state(name=ConnectionState.ERROR)
                    else:
                        self.change_state(name=ConnectionState.CONNECTED)
                        redo = True
            finally:
                self.__connecting -= 1
        return redo

    @property  # Override
    def channel(self) -> Channel:
        if self._channel is None:
            self.__reconnect()
        return self._channel

    @property  # Override
    def opened(self) -> bool:
        return self.__running

    # Override
    def start(self):
        self.__running = True
        super().start()

    # Override
    def stop(self):
        self.__running = False
        super().stop()

    # Override
    def _receive(self, max_len: int) -> (bytes, tuple):
        data, remote = super()._receive(max_len=max_len)
        if data is None and remote is None and self._channel is None and self.__reconnect():
            # try again
            data, remote = super()._receive(max_len=max_len)
        return data, remote

    # Override
    def send(self, data: bytes, target: Optional[tuple] = None) -> int:
        sent = super().send(data=data, target=target)
        if sent == -1 and self._channel is None and self.__reconnect():
            # try again
            sent = super().send(data=data, target=target)
        return sent

    # Override
    def exit_state(self, state: ConnectionState, ctx: StateMachine):
        super().exit_state(state=state, ctx=ctx)
        current = ctx.current_state
        if current == ConnectionState.EXPIRED:
            try:
                self.heartbeat()
            except socket.error as error:
                print('[NET] failed to heartbeat: %s' % error)

    def heartbeat(self):
        # send 'PING'
        pass



# -*- coding: utf-8 -*-
#
#   TCP: Transmission Control Protocol
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import weakref
from abc import ABC

from ..fsm import Context, BaseTransition, BaseState, BaseMachine

from .connection import Connection


"""
    Finite States:

            //===============\\          (Sent)          //==============\\
            ||               || -----------------------> ||              ||
            ||    Default    ||                          ||  Connecting  ||
            || (Not Connect) || <----------------------- ||              ||
            \\===============//         (Timeout)        \\==============//
                                                              |       |
            //===============\\                               |       |
            ||               || <-----------------------------+       |
            ||     Error     ||          (Error)                 (Received)
            ||               || <-----------------------------+       |
            \\===============//                               |       |
                A       A                                     |       |
                |       |            //===========\\          |       |
                (Error) +----------- ||           ||          |       |
                |                    ||  Expired  || <--------+       |
                |       +----------> ||           ||          |       |
                |       |            \\===========//          |       |
                |       (Timeout)           |         (Timeout)       |
                |       |                   |                 |       V
            //===============\\     (Sent)  |            //==============\\
            ||               || <-----------+            ||              ||
            ||  Maintaining  ||                          ||  Connected   ||
            ||               || -----------------------> ||              ||
            \\===============//       (Received)         \\==============//
"""


class StateMachine(BaseMachine, Context):

    def __init__(self, connection: Connection):
        super().__init__(default=ConnectionState.DEFAULT)
        self.__connection = weakref.ref(connection)
        # init states
        builder = self._create_state_builder()
        self.__set_state(state=builder.get_default_state())
        self.__set_state(state=builder.get_connecting_state())
        self.__set_state(state=builder.get_connected_state())
        self.__set_state(state=builder.get_expired_state())
        self.__set_state(state=builder.get_maintaining_state())
        self.__set_state(state=builder.get_error_state())

    # noinspection PyMethodMayBeStatic
    def _create_state_builder(self):
        return StateBuilder(transition_builder=TransitionBuilder())

    def __set_state(self, state):
        self.add_state(name=state.name, state=state)

    @property
    def context(self) -> Context:
        return self

    @property
    def connection(self) -> Connection:
        return self.__connection()


class StateTransition(BaseTransition[StateMachine], ABC):
    pass


class ConnectionState(BaseState[StateMachine, StateTransition]):
    """
        Connection State
        ~~~~~~~~~~~~~~~~

        Defined for indicating connection status

            DEFAULT     - 'initialized', or sent timeout
            CONNECTING  - sent 'PING', waiting for response
            CONNECTED   - got response recently
            EXPIRED     - long time, needs maintaining (still connected)
            MAINTAINING - sent 'PING', waiting for response
            ERROR       - long long time no response, connection lost
    """

    DEFAULT = 'default'
    CONNECTING = 'connecting'
    CONNECTED = 'connected'
    MAINTAINING = 'maintaining'
    EXPIRED = 'expired'
    ERROR = 'error'

    def __init__(self, name: str):
        super(ConnectionState, self).__init__()
        self.__name = name

    @property
    def name(self) -> str:
        return self.__name

    def __str__(self) -> str:
        return self.__name

    def __repr__(self) -> str:
        return self.__name

    def __eq__(self, other) -> bool:
        if self is other:
            return True
        elif isinstance(other, ConnectionState):
            return self.__name == other.name
        elif isinstance(other, str):
            return self.__name == other
        else:
            return False

    def __ne__(self, other) -> bool:
        if self is other:
            return False
        elif isinstance(other, ConnectionState):
            return self.__name != other.name
        elif isinstance(other, str):
            return self.__name != other
        else:
            return True

    def on_enter(self, ctx: StateMachine):
        pass

    def on_exit(self, ctx: StateMachine):
        pass

    def on_pause(self, ctx: StateMachine):
        pass

    def on_resume(self, ctx: StateMachine):
        pass


#
#   Builders
#
class StateBuilder:

    def __init__(self, transition_builder):
        super().__init__()
        self.__builder = transition_builder

    # noinspection PyMethodMayBeStatic
    def get_named_state(self, name: str) -> ConnectionState:
        return ConnectionState(name=name)

    # Connection not started yet
    def get_default_state(self) -> ConnectionState:
        builder = self.__builder
        # assert isinstance(builder, TransitionBuilder)
        state = self.get_named_state(name=ConnectionState.DEFAULT)
        # Default -> Connecting
        state.add_transition(transition=builder.get_default_connecting_transition())
        return state

    # Connection started, not connected yet
    def get_connecting_state(self) -> ConnectionState:
        builder = self.__builder
        # assert isinstance(builder, TransitionBuilder)
        state = self.get_named_state(name=ConnectionState.CONNECTING)
        # Connecting -> Connected
        state.add_transition(transition=builder.get_connecting_connected_transition())
        # Connecting -> Default
        state.add_transition(transition=builder.get_connecting_default_transition())
        return state

    # Normal state of connection
    def get_connected_state(self) -> ConnectionState:
        builder = self.__builder
        # assert isinstance(builder, TransitionBuilder)
        state = self.get_named_state(name=ConnectionState.CONNECTED)
        # Connected -> Expired
        state.add_transition(transition=builder.get_connected_expired_transition())
        # Connected -> Error
        state.add_transition(transition=builder.get_connected_error_transition())
        return state

    # Long time no response, need maintaining
    def get_expired_state(self) -> ConnectionState:
        builder = self.__builder
        # assert isinstance(builder, TransitionBuilder)
        state = self.get_named_state(name=ConnectionState.EXPIRED)
        # Expired -> Maintaining
        state.add_transition(transition=builder.get_expired_maintaining_transition())
        # Expired -> Error
        state.add_transition(transition=builder.get_expired_error_transition())
        return state

    # Heartbeat sent, waiting response
    def get_maintaining_state(self) -> ConnectionState:
        builder = self.__builder
        # assert isinstance(builder, TransitionBuilder)
        state = self.get_named_state(name=ConnectionState.MAINTAINING)
        # Maintaining -> Connected
        state.add_transition(transition=builder.get_maintaining_connected_transition())
        # Maintaining -> Expired
        state.add_transition(transition=builder.get_maintaining_expired_transition())
        # Maintaining -> Error
        state.add_transition(transition=builder.get_maintaining_error_transition())
        return state

    # Connection lost
    def get_error_state(self) -> ConnectionState:
        builder = self.__builder
        # assert isinstance(builder, TransitionBuilder)
        state = self.get_named_state(name=ConnectionState.ERROR)
        # Error -> Default
        state.add_transition(transition=builder.get_error_default_transition())
        return state


class TransitionBuilder:

    # noinspection PyMethodMayBeStatic
    def get_default_connecting_transition(self):
        return DefaultConnectingTransition(target=ConnectionState.CONNECTING)

    # Connecting

    # noinspection PyMethodMayBeStatic
    def get_connecting_connected_transition(self):
        return ConnectingConnectedTransition(target=ConnectionState.CONNECTED)

    # noinspection PyMethodMayBeStatic
    def get_connecting_default_transition(self):
        return ConnectingDefaultTransition(target=ConnectionState.DEFAULT)

    # Connected

    # noinspection PyMethodMayBeStatic
    def get_connected_expired_transition(self):
        return ConnectedExpiredTransition(target=ConnectionState.EXPIRED)

    # noinspection PyMethodMayBeStatic
    def get_connected_error_transition(self):
        return ConnectedErrorTransition(target=ConnectionState.ERROR)

    # Expired

    # noinspection PyMethodMayBeStatic
    def get_expired_maintaining_transition(self):
        return ExpiredMaintainingTransition(target=ConnectionState.MAINTAINING)

    # noinspection PyMethodMayBeStatic
    def get_expired_error_transition(self):
        return ExpiredErrorTransition(target=ConnectionState.ERROR)

    # Maintaining

    # noinspection PyMethodMayBeStatic
    def get_maintaining_connected_transition(self):
        return MaintainingConnectedTransition(target=ConnectionState.CONNECTED)

    # noinspection PyMethodMayBeStatic
    def get_maintaining_expired_transition(self):
        return MaintainingExpiredTransition(target=ConnectionState.EXPIRED)

    # noinspection PyMethodMayBeStatic
    def get_maintaining_error_transition(self):
        return MaintainingErrorTransition(target=ConnectionState.ERROR)

    # Error

    # noinspection PyMethodMayBeStatic
    def get_error_default_transition(self) -> StateTransition:
        return ErrorDefaultTransition(target=ConnectionState.DEFAULT)


#
#   Transitions
#

class DefaultConnectingTransition(StateTransition):
    """ Default -> Connecting """

    def evaluate(self, ctx: StateMachine) -> bool:
        conn = ctx.connection
        # connection started? change state to 'connecting'
        return conn is not None and conn.opened


class ConnectingConnectedTransition(StateTransition):
    """ Connecting -> Connected """

    def evaluate(self, ctx: StateMachine) -> bool:
        conn = ctx.connection
        # connection connected, change state to 'connected'
        return conn is not None and conn.opened and conn.connected


class ConnectingDefaultTransition(StateTransition):
    """ Connecting -> Default """

    def evaluate(self, ctx: StateMachine) -> bool:
        conn = ctx.connection
        # connection stopped, change state to 'not_connect'
        return conn is None or not conn.opened


class ConnectedExpiredTransition(StateTransition):
    """ Connected -> Expired """

    def evaluate(self, ctx: StateMachine) -> bool:
        conn = ctx.connection
        from .base_conn import BaseConnection
        assert isinstance(conn, BaseConnection), 'connection error: %s' % conn
        # connection still alive, but
        # long time no response, change state to 'maintain_expired'
        return conn is not None and conn.opened and conn.connected and not conn.received_recently


class ConnectedErrorTransition(StateTransition):
    """ Connected -> Error """

    def evaluate(self, ctx: StateMachine) -> bool:
        conn = ctx.connection
        # connection lost, change state to 'error
        return conn is None or not conn.opened


class ExpiredMaintainingTransition(StateTransition):
    """ Expired -> Maintaining """

    def evaluate(self, ctx: StateMachine) -> bool:
        conn = ctx.connection
        from .base_conn import BaseConnection
        assert isinstance(conn, BaseConnection), 'connection error: %s' % conn
        # connection still alive, and
        # sent recently, change state to 'maintaining'
        return conn is not None and conn.opened and conn.connected and conn.sent_recently


class ExpiredErrorTransition(StateTransition):
    """ Expired -> Error """

    def evaluate(self, ctx: StateMachine) -> bool:
        conn = ctx.connection
        # connection lost, change state to 'error
        return conn is None or not conn.opened


class MaintainingConnectedTransition(StateTransition):
    """ Maintaining -> Connected """

    def evaluate(self, ctx: StateMachine) -> bool:
        conn = ctx.connection
        from .base_conn import BaseConnection
        assert isinstance(conn, BaseConnection), 'connection error: %s' % conn
        # connection still alive, and
        # received recently, change state to 'connected
        return conn is not None and conn.opened and conn.connected and conn.received_recently


class MaintainingExpiredTransition(StateTransition):
    """ Maintaining -> Expired """

    def evaluate(self, ctx: StateMachine) -> bool:
        conn = ctx.connection
        from .base_conn import BaseConnection
        assert isinstance(conn, BaseConnection), 'connection error: %s' % conn
        # connection still alive, but
        # long time no sending, change state to 'maintain_expired'
        return conn is not None and conn.opened and conn.connected and not conn.sent_recently


class MaintainingErrorTransition(StateTransition):
    """ Maintaining -> Error """

    def evaluate(self, ctx: StateMachine) -> bool:
        conn = ctx.connection
        from .base_conn import BaseConnection
        assert isinstance(conn, BaseConnection), 'connection error: %s' % conn
        # connection lost, or
        # long long time no response, change state to 'error'
        return conn is None or not conn.opened or conn.long_time_not_received


class ErrorDefaultTransition(StateTransition):
    """ Error -> Default """

    def evaluate(self, ctx: StateMachine) -> bool:
        conn = ctx.connection
        # connection reset, change state to 'not_connect'
        return conn is not None and conn.opened



# -*- coding: utf-8 -*-
#
#   TCP: Transmission Control Protocol
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import socket
import time
import weakref
from typing import Optional

from ..fsm import Delegate as StateDelegate

from .channel import Channel
from .connection import Connection, Delegate as ConnectionDelegate
from .state import ConnectionState, StateMachine


class BaseConnection(Connection, StateDelegate):

    """
        Maximum Segment Size
        ~~~~~~~~~~~~~~~~~~~~
        Buffer size for receiving package

        MTU        : 1500 bytes (excludes 14 bytes ethernet header & 4 bytes FCS)
        IP header  :   20 bytes
        TCP header :   20 bytes
        UDP header :    8 bytes
    """
    MSS = 1472  # 1500 - 20 - 8

    EXPIRES = 16  # seconds

    def __init__(self, remote: Optional[tuple], local: Optional[tuple], channel: Optional[Channel]):
        super().__init__()
        self._channel = channel
        self._remote_address = remote
        self._local_address = local
        # active times
        self.__last_sent_time = 0
        self.__last_received_time = 0
        # Connection Delegate
        self.__delegate: Optional[weakref.ReferenceType] = None
        # Finite State Machine
        self.__fsm = self._create_state_machine()

    def _create_state_machine(self) -> StateMachine:
        fsm = StateMachine(connection=self)
        fsm.delegate = self
        return fsm

    @property
    def delegate(self) -> ConnectionDelegate:
        if self.__delegate is not None:
            return self.__delegate()

    @delegate.setter
    def delegate(self, handler: ConnectionDelegate):
        if handler is None:
            self.__delegate = None
        else:
            self.__delegate = weakref.ref(handler)

    def __eq__(self, other) -> bool:
        if self is other:
            return True
        elif isinstance(other, Connection):
            return self.remote_address == other.remote_address and self.local_address == other.local_address

    def __ne__(self, other) -> bool:
        if self is other:
            return False
        elif isinstance(other, Connection):
            return self.remote_address != other.remote_address or self.local_address != other.local_address
        else:
            return True

    def __hash__(self) -> int:
        local = self.local_address
        remote = self.remote_address
        if remote is None:
            assert local is not None, 'both local & remote addresses are empty'
            return hash(local)
        # same algorithm as Pair::hashCode()
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # remote's hashCode is multiplied by an arbitrary prime number (13)
        # in order to make sure there is a difference in the hashCode between
        # these two parameters:
        #     remote: a  local: aa
        #     local: aa  remote: a
        if local is None:
            return hash(remote) * 13
        else:
            return hash(remote) * 13 + hash(local)

    @property  # Override
    def local_address(self) -> Optional[tuple]:  # (str, int)
        sock = self._channel
        if sock is not None and sock.bound:
            return sock.local_address
        else:
            return self._local_address

    @property  # Override
    def remote_address(self) -> Optional[tuple]:  # (str, int)
        sock = self._channel
        if sock is not None and sock.connected:
            return sock.remote_address
        else:
            return self._remote_address

    @property
    def sent_recently(self) -> bool:
        now = time.time()
        return now < self.__last_sent_time + self.EXPIRES

    @property
    def received_recently(self) -> bool:
        now = time.time()
        return now < self.__last_received_time + self.EXPIRES

    @property
    def long_time_not_received(self) -> bool:
        now = time.time()
        return now > self.__last_received_time + (self.EXPIRES << 4)

    @property  # Override
    def opened(self) -> bool:
        sock = self.channel
        return sock is not None and sock.opened

    @property  # Override
    def bound(self) -> bool:
        sock = self.channel
        return sock is not None and sock.bound

    @property  # Override
    def connected(self) -> bool:
        sock = self.channel
        return sock is not None and sock.connected

    # Override
    def close(self):
        try:
            sock = self._channel
            if sock is not None and sock.opened:
                sock.close()
        except socket.error as error:
            print('[NET] failed to close socket: %s' % error)
        finally:
            self._channel = None
            self.change_state(name=ConnectionState.DEFAULT)

    @property
    def channel(self) -> Channel:
        return self._channel

    def _receive(self, max_len: int) -> (bytes, tuple):
        sock = self.channel
        if sock is None or not sock.opened:
            raise socket.error('connection lost: %s' % sock)
        try:
            data, remote = sock.receive(max_len=max_len)
            if data is not None and len(data) > 0:
                self.__last_received_time = time.time()
            return data, remote
        except socket.error as error:
            print('[NET] failed to receive data: %s' % error)
            self.close()
            self.change_state(name=ConnectionState.ERROR)
            raise error

    # Override
    def send(self, data: bytes, target: Optional[tuple] = None) -> int:
        sock = self.channel
        if sock is None or not sock.opened:
            raise socket.error('connection lost: %s' % sock)
        try:
            sent = sock.send(data=data, target=target)
            if sent != -1:
                self.__last_sent_time = time.time()
            return sent
        except socket.error as error:
            print('[NET] failed to send data: %s' % error)
            self.close()
            self.change_state(name=ConnectionState.ERROR)
            raise error

    #
    #   States
    #

    def change_state(self, name: str):
        state = self.__fsm.get_state(name=name)
        if state != self.__fsm.current_state:
            self.__fsm.change_state(state=state)

    @property  # Override
    def state(self) -> ConnectionState:
        return self.__fsm.current_state

    #
    #   Ticker
    #

    # Override
    def tick(self):
        self.__fsm.tick()
        if self.opened:
            # try to receive data when connection open
            try:
                self._process()
            except socket.error as error:
                print('[NET] failed to process: %s' % error)

    def _process(self):
        delegate = self.delegate
        if delegate is None:
            return
        # receiving
        data, remote = self._receive(max_len=self.MSS)
        if data is None or len(data) == 0:
            return
        # callback
        delegate.connection_data_received(connection=self, remote=remote, wrapper=None, payload=data)

    def start(self):
        self.__fsm.start()

    def stop(self):
        self.close()
        self.__fsm.stop()

    #
    #   StateDelegate
    #

    # Override
    def enter_state(self, state: ConnectionState, ctx: StateMachine):
        pass

    # Override
    def exit_state(self, state: ConnectionState, ctx: StateMachine):
        current = ctx.current_state
        if current == ConnectionState.CONNECTED:
            if state != ConnectionState.MAINTAINING:
                # change state to 'connected', reset times to just expired
                timestamp = time.time() - self.EXPIRES - 1
                self.__last_sent_time = timestamp
                self.__last_received_time = timestamp
        # callback
        delegate = self.delegate
        if delegate is not None:
            delegate.connection_state_changed(connection=self, previous=state, current=current)

    # Override
    def pause_state(self, state: ConnectionState, ctx: StateMachine):
        pass

    # Override
    def resume_state(self, state: ConnectionState, ctx: StateMachine):
        pass



# -*- coding: utf-8 -*-
#
#   TCP: Transmission Control Protocol
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import socket
from abc import ABC, abstractmethod
from typing import Optional, Set

from .connection import Connection


"""
    Topology
    ~~~~~~~~

                            +---------------+
                            |      APP      |
                            +---------------+
                                |       A
                                |       |  (filter)
                                V       |
            +-----------------------------------------------+
            |                                               |
            |     +----------+     HUB     +----------+     |
            |     |  socket  |             |  socket  |     |
            +-----+----------+-------------+----------+-----+
                     |    A                   |  |  A
                     |    |   (connections)   |  |  |
                     |    |    (+channels)    |  |  |
                     |    |                   |  |  |
            ~~~~~~~~~|~~~~|~~~~~~~~~~~~~~~~~~~|~~|~~|~~~~~~~~
            ~~~~~~~~~|~~~~|~~~~~~~~~~~~~~~~~~~|~~|~~|~~~~~~~~
                     |    |                   |  |  |
                     V    |                   V  V  |
"""


class Hub(ABC):

    @abstractmethod
    def send(self, data: bytes, source: Optional[tuple], destination: tuple) -> bool:
        """
        Send data from source to destination

        :param data:        payload
        :param source:      from address (local);
                            if it's None, send via any connection connected to destination
        :param destination: to address (remote)
        :return: False on error
        """
        raise NotImplemented

    @abstractmethod
    def get_connection(self, remote: Optional[tuple], local: Optional[tuple]) -> Optional[Connection]:
        """
        Get connection if already exists

        :param remote: remote address
        :param local:  local address
        :return: None on connection not found
        """
        raise NotImplemented

    @abstractmethod
    def connect(self, remote: Optional[tuple], local: Optional[tuple]) -> Optional[Connection]:
        """
        Get/create connection

        :param remote: remote address
        :param local:  local address
        :return: None on error
        """
        raise NotImplemented

    @abstractmethod
    def disconnect(self, remote: Optional[tuple], local: Optional[tuple]):
        """
        Close connection

        :param remote: remote address
        :param local:  local address
        """
        raise NotImplemented

    #
    #   Local Address
    #

    @classmethod
    def host_name(cls) -> str:
        return socket.gethostname()

    @classmethod
    def addr_info(cls):  # -> List[Tuple[Union[AddressFamily, int], Union[SocketKind, int], int, str, Tuple[Any, ...]]]
        host = socket.gethostname()
        if host is not None:
            return socket.getaddrinfo(host, None)

    @classmethod
    def inet_addresses(cls) -> Set[str]:
        addresses = set()
        info = cls.addr_info()
        for item in info:
            addresses.add(item[4][0])
        return addresses

    @classmethod
    def inet_address(cls) -> Optional[str]:
        # get from addr info
        info = cls.addr_info()
        for item in info:
            ip = item[4][0]
            if ':' not in ip and '127.0.0.1' != ip:
                return ip
        # get from UDP socket
        sock = None
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            remote = ('8.8.8.8', 8888)
            sock.connect(remote)
            ip = sock.getsockname()[0]
        finally:
            if sock is not None:
                sock.close()
        return ip



# -*- coding: utf-8 -*-
#
#   TCP: Transmission Control Protocol
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import socket
from abc import ABC, abstractmethod
from typing import Optional

from .channel import Channel


class BaseChannel(Channel, ABC):

    def __init__(self, sock: Optional[socket.socket] = None,
                 remote: Optional[tuple] = None, local: Optional[tuple] = None,
                 blocking: bool = True, reuse: bool = True):
        super().__init__()
        if sock is None:
            # BaseChannel(remote=remote, local=local, blocking=blocking, reuse=reuse),
            self._blocking = blocking
            self._reuse = reuse
            # setup inner socket
            self._sock = None
            sock = self._setup_socket()
            # bind to local address
            if local is not None:
                sock.bind(local)
            # connect to remote address
            if remote is not None:
                sock.connect(remote)
        else:
            # BaseChannel(sock=sock)
            self._blocking = sock.getblocking()
            self._reuse = getattr(sock, 'SO_REUSEPORT', 0)
            self._sock = sock

    @abstractmethod
    def _setup_socket(self) -> socket.socket:
        """ create socket (blocking, reuse) """
        raise NotImplemented

    # Override
    def configure_blocking(self, blocking: bool):
        self._blocking = blocking
        sock = self._sock
        if sock is None:
            sock = self._setup_socket()
        else:
            sock.setblocking(blocking)
        return sock

    @property  # Override
    def blocking(self) -> bool:
        sock = self._sock
        if sock is None:
            return self._blocking
        else:
            return sock.getblocking()

    @property  # Override
    def opened(self) -> bool:
        sock = self._sock
        return sock is not None and not getattr(sock, '_closed', False)

    @property  # Override
    def connected(self) -> bool:
        try:
            return self.remote_address is not None
        except socket.error:
            return False

    @property  # Override
    def bound(self) -> bool:
        try:
            return self.local_address is not None
        except socket.error:
            return False

    @property  # Override
    def local_address(self) -> Optional[tuple]:
        sock = self._sock
        if sock is None:
            return None
        else:
            # assert isinstance(sock, socket.socket)
            return sock.getsockname()

    @property  # Override
    def remote_address(self) -> Optional[tuple]:
        sock = self._sock
        if sock is None:
            return None
        try:
            # assert isinstance(sock, socket.socket)
            return sock.getpeername()
        except socket.error as error:
            if error.strerror == 'Socket is not connected':
                return None
            else:
                raise error

    # Override
    def bind(self, address: Optional[tuple] = None, host: Optional[str] = '0.0.0.0', port: Optional[int] = 0):
        if address is None:
            address = (host, port)
        sock = self._setup_socket()
        # assert isinstance(sock, socket.socket)
        sock.bind(address)

    # Override
    def connect(self, address: Optional[tuple] = None, host: Optional[str] = '127.0.0.1', port: Optional[int] = 0):
        if address is None:
            address = (host, port)
        sock = self._setup_socket()
        # assert isinstance(sock, socket.socket)
        sock.connect(address)

    # Override
    def disconnect(self):
        sock = self._sock
        self.close()
        return sock

    # Override
    def close(self):
        sock = self._sock
        if sock is not None and not getattr(sock, '_closed', False):
            # sock.shutdown(socket.SHUT_RDWR)
            sock.close()
        self._sock = None

    #
    #   Input/Output
    #

    # Override
    def read(self, max_len: int) -> Optional[bytes]:
        # check socket
        sock = self._sock
        if sock is None:
            raise socket.error('socket lost, cannot read data')
        # try to receive data
        try:
            data = sock.recv(max_len)
        except socket.error as error:
            # the socket will raise 'Resource temporarily unavailable'
            # when received nothing in non-blocking mode,
            # here we should ignore this exception.
            if not self._blocking:
                if error.strerror == 'Resource temporarily unavailable':
                    # received nothing
                    return None
            raise error
        # check data
        if data is None or len(data) == 0:
            # in blocking mode, the socket will wait until received something,
            # but if timeout was set, it will return None too, it's normal;
            # otherwise, we know the connection was lost.
            if sock.gettimeout() is None:  # and self._blocking
                raise socket.error('remote peer reset socket')
        return data

    # Override
    def write(self, data: bytes) -> int:
        sock = self._sock
        if sock is None:
            raise socket.error('socket lost, cannot write data: %d byte(s)' % len(data))
        # sock.sendall(data)
        sent = 0
        rest = len(data)
        while rest > 0:  # and not getattr(sock, '_closed', False):
            cnt = sock.send(data)
            if cnt > 0:
                sent += cnt
                rest -= cnt
                data = data[cnt:]
        return sent
